@@ -16,6 +16,7 @@ package e2e
 
 import (
 	"bytes"
+	"compress/gzip"
 	"context"
 	"encoding/json"
 	"fmt"
@@ -27,7 +28,7 @@ import (
 	"testing"
 	"time"
 
-	"k8s.io/api/core/v1"
+	v1 "k8s.io/api/core/v1"
 	apierrors "k8s.io/apimachinery/pkg/api/errors"
 	"k8s.io/apimachinery/pkg/api/resource"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
@@ -163,7 +164,7 @@ func testPromResourceUpdate(t *testing.T) {
 		t.Fatal(err)
 	}
 
-	pods, err := framework.KubeClient.Core().Pods(ns).List(prometheus.ListOptions(name))
+	pods, err := framework.KubeClient.CoreV1().Pods(ns).List(prometheus.ListOptions(name))
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -184,7 +185,7 @@ func testPromResourceUpdate(t *testing.T) {
 	}
 
 	err = wait.Poll(5*time.Second, 2*time.Minute, func() (bool, error) {
-		pods, err := framework.KubeClient.Core().Pods(ns).List(prometheus.ListOptions(name))
+		pods, err := framework.KubeClient.CoreV1().Pods(ns).List(prometheus.ListOptions(name))
 		if err != nil {
 			return false, err
 		}
@@ -244,7 +245,7 @@ func testPromStorageUpdate(t *testing.T) {
 	}
 
 	err = wait.Poll(5*time.Second, 2*time.Minute, func() (bool, error) {
-		pods, err := framework.KubeClient.Core().Pods(ns).List(prometheus.ListOptions(name))
+		pods, err := framework.KubeClient.CoreV1().Pods(ns).List(prometheus.ListOptions(name))
 		if err != nil {
 			return false, err
 		}
@@ -290,13 +291,19 @@ scrape_configs:
         - 111.111.111.111:9090
 `
 
+	var bufOne bytes.Buffer
+	if err := gzipConfig(&bufOne, []byte(firstConfig)); err != nil {
+		t.Fatal(err)
+	}
+	firstConfigCompressed := bufOne.Bytes()
+
 	cfg := &v1.Secret{
 		ObjectMeta: metav1.ObjectMeta{
 			Name: fmt.Sprintf("prometheus-%s", name),
 		},
 		Data: map[string][]byte{
-			"prometheus.yaml": []byte(firstConfig),
-			"configmaps.json": []byte("{}"),
+			"prometheus.yaml.gz": firstConfigCompressed,
+			"configmaps.json":    []byte("{}"),
 		},
 	}
 
@@ -332,12 +339,18 @@ scrape_configs:
         - 111.111.111.112:9090
 `
 
+	var bufTwo bytes.Buffer
+	if err := gzipConfig(&bufTwo, []byte(secondConfig)); err != nil {
+		t.Fatal(err)
+	}
+	secondConfigCompressed := bufTwo.Bytes()
+
 	cfg, err := framework.KubeClient.CoreV1().Secrets(ns).Get(cfg.Name, metav1.GetOptions{})
 	if err != nil {
 		t.Fatal(errors.Wrap(err, "could not retrieve previous secret"))
 	}
 
-	cfg.Data["prometheus.yaml"] = []byte(secondConfig)
+	cfg.Data["prometheus.yaml.gz"] = secondConfigCompressed
 	if _, err := framework.KubeClient.CoreV1().Secrets(ns).Update(cfg); err != nil {
 		t.Fatal(err)
 	}
@@ -718,6 +731,53 @@ func testPromRulesExceedingConfigMapLimit(t *testing.T) {
 	}
 }
 
+func testPromRulesMustBeAnnotated(t *testing.T) {
+	t.Parallel()
+
+	ctx := framework.NewTestCtx(t)
+	defer ctx.Cleanup(t)
+	ns := ctx.CreateNamespace(t, framework.KubeClient)
+	ctx.SetupPrometheusRBAC(t, ns, framework.KubeClient)
+
+	name := "admission"
+	admissionAlert := "admissionAlert"
+
+	_, err := framework.MakeAndCreateFiringRule(ns, name, admissionAlert)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	rule, err := framework.GetRule(ns, name)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	val, ok := rule.Annotations["prometheus-operator-validated"]
+	if !ok {
+		t.Fatal("Expected prometheusrule to be annotated")
+	}
+	if val != "true" {
+		t.Fatal("Expected prometheusrule annotation to be 'true'")
+	}
+}
+
+func testInvalidRulesAreRejected(t *testing.T) {
+	t.Parallel()
+
+	ctx := framework.NewTestCtx(t)
+	defer ctx.Cleanup(t)
+	ns := ctx.CreateNamespace(t, framework.KubeClient)
+	ctx.SetupPrometheusRBAC(t, ns, framework.KubeClient)
+
+	name := "admission"
+	admissionAlert := "admissionAlert"
+
+	_, err := framework.MakeAndCreateInvalidRule(ns, name, admissionAlert)
+	if err == nil {
+		t.Fatal("Expected invalid prometheusrule to be rejected")
+	}
+}
+
 // generateHugePrometheusRule returns a Prometheus rule instance that would fill
 // more than half of the space of a Kubernetes ConfigMap.
 func generateHugePrometheusRule(ns, identifier string) *monitoringv1.PrometheusRule {
@@ -1207,70 +1267,62 @@ func testThanos(t *testing.T) {
 	ns := ctx.CreateNamespace(t, framework.KubeClient)
 	ctx.SetupPrometheusRBAC(t, ns, framework.KubeClient)
 
-	peerServiceName := "thanos-peers"
-	querierServiceName := "thanos-querier"
-	basicPrometheus := framework.MakeBasicPrometheus(ns, "basic-prometheus", "test-group", 1)
-	peerServiceDNS := fmt.Sprintf("%s.%s.svc:10900", peerServiceName, ns)
-	version := "v0.1.0"
-	basicPrometheus.Spec.Thanos = &monitoringv1.ThanosSpec{
-		Peers:   &peerServiceDNS,
+	version := prometheus.DefaultThanosVersion
+
+	prom := framework.MakeBasicPrometheus(ns, "basic-prometheus", "test-group", 1)
+	prom.Spec.Replicas = proto.Int32(2)
+	prom.Spec.Thanos = &monitoringv1.ThanosSpec{
 		Version: &version,
 	}
-	basicPrometheus.Spec.PodMetadata = &metav1.ObjectMeta{
-		CreationTimestamp: metav1.Now(),
-		Labels: map[string]string{
-			"thanos-peer": "true",
-		},
-	}
-	replicas := int32(2)
-	basicPrometheus.Spec.Replicas = &replicas
-	pservice := framework.MakePrometheusService(basicPrometheus.Name, "test-group", v1.ServiceTypeClusterIP)
-	tservice := framework.MakeThanosService(peerServiceName)
-	qservice := framework.MakeThanosQuerierService(querierServiceName)
-	s := framework.MakeBasicServiceMonitor("test-group")
-	thanosQuerier, err := testFramework.MakeDeployment("../../example/thanos/querier-deployment.yaml")
-	if err != nil {
-		t.Fatal("Making deployment failed: ", err)
-	}
-	querierArgs := []string{
-		"query",
-		"--log.level=debug",
-		"--query.replica-label=prometheus_replica",
-		fmt.Sprintf("--cluster.peers=%s", peerServiceDNS),
-	}
-	log.Println("setting up querier with args: ", querierArgs)
-	thanosQuerier.Spec.Template.Spec.Containers[0].Args = querierArgs
-
-	if err := testFramework.CreateDeployment(framework.KubeClient, ns, thanosQuerier); err != nil {
-		t.Fatal("Creating Thanos querier failed: ", err)
-	}
-
-	if _, err := testFramework.CreateServiceAndWaitUntilReady(framework.KubeClient, ns, qservice); err != nil {
-		t.Fatal("Creating Thanos querier service failed: ", err)
-	}
-
-	if _, err := framework.MonClientV1.ServiceMonitors(ns).Create(s); err != nil {
-		t.Fatal("Creating ServiceMonitor failed: ", err)
-	}
-
-	if _, err := framework.KubeClient.CoreV1().Services(ns).Create(pservice); err != nil {
-		t.Fatal("Creating prometheus service failed: ", err)
-	}
-
-	if _, err := framework.MonClientV1.Prometheuses(ns).Create(basicPrometheus); err != nil {
+	if _, err := framework.CreatePrometheusAndWaitUntilReady(ns, prom); err != nil {
 		t.Fatal("Creating prometheus failed: ", err)
 	}
 
-	if _, err := framework.KubeClient.CoreV1().Services(ns).Create(tservice); err != nil {
+	promSvc := framework.MakePrometheusService(prom.Name, "test-group", v1.ServiceTypeClusterIP)
+	if _, err := framework.KubeClient.CoreV1().Services(ns).Create(promSvc); err != nil {
 		t.Fatal("Creating prometheus service failed: ", err)
 	}
 
+	svcMon := framework.MakeBasicServiceMonitor("test-group")
+	if _, err := framework.MonClientV1.ServiceMonitors(ns).Create(svcMon); err != nil {
+		t.Fatal("Creating ServiceMonitor failed: ", err)
+	}
+
+	qryDep, err := testFramework.MakeDeployment("../../example/thanos/query-deployment.yaml")
+	if err != nil {
+		t.Fatal("Making thanos query deployment failed: ", err)
+	}
+	// override image
+	qryImage := "quay.io/thanos/thanos:" + version
+	t.Log("setting up query with image: ", qryImage)
+	qryDep.Spec.Template.Spec.Containers[0].Image = qryImage
+	// override args
+	qryArgs := []string{
+		"query",
+		"--log.level=debug",
+		"--query.replica-label=prometheus_replica",
+		fmt.Sprintf("--store=dnssrv+_grpc._tcp.prometheus-operated.%s.svc.cluster.local", ns),
+	}
+	t.Log("setting up query with args: ", qryArgs)
+	qryDep.Spec.Template.Spec.Containers[0].Args = qryArgs
+	if err := testFramework.CreateDeployment(framework.KubeClient, ns, qryDep); err != nil {
+		t.Fatal("Creating Thanos query deployment failed: ", err)
+	}
+
+	qrySvc := framework.MakeThanosQuerierService(qryDep.Name)
+	if _, err := testFramework.CreateServiceAndWaitUntilReady(framework.KubeClient, ns, qrySvc); err != nil {
+		t.Fatal("Creating Thanos query service failed: ", err)
+	}
+
 	err = wait.Poll(5*time.Second, 5*time.Minute, func() (bool, error) {
 		proxyGet := framework.KubeClient.CoreV1().Services(ns).ProxyGet
-		request := proxyGet("http", querierServiceName, "http-query", "/api/v1/query", map[string]string{"query": "prometheus_build_info", "dedup": "false"})
+		request := proxyGet("http", qrySvc.Name, "http-query", "/api/v1/query", map[string]string{
+			"query": "prometheus_build_info",
+			"dedup": "false",
+		})
 		b, err := request.DoRaw()
 		if err != nil {
-			log.Println(fmt.Sprintf("Error performing request against Thanos querier: %v\n\nretrying...", err))
+			t.Logf("Error performing request against Thanos query: %v\n\nretrying...", err)
 			return false, nil
 		}
 
@@ -1291,13 +1343,13 @@ func testThanos(t *testing.T) {
 		// Prometheus replicas by both replicas.
 		expected := 4
 		if result != expected {
-			log.Printf("Unexpected number of results from query. Got %d, expected %d. retrying...\n", result, expected)
+			t.Logf("Unexpected number of results from query. Got %d, expected %d. retrying...", result, expected)
 			return false, nil
 		}
 		return true, nil
 	})
 	if err != nil {
-		t.Fatal("Failed to get correct result from Thanos querier: ", err)
+		t.Fatal("Failed to get correct result from Thanos query: ", err)
 	}
 }
 
@@ -1435,7 +1487,7 @@ func testOperatorNSScope(t *testing.T) {
 		}
 
 		// Prometheus Operator only watches single namespace mainNS, not arbitraryNS.
-		err := framework.CreatePrometheusOperator(operatorNS, *opImage, []string{mainNS})
+		_, err := framework.CreatePrometheusOperator(operatorNS, *opImage, []string{mainNS}, nil, false)
 		if err != nil {
 			t.Fatal(err)
 		}
@@ -1505,7 +1557,7 @@ func testOperatorNSScope(t *testing.T) {
 		}
 
 		// Prometheus Operator only watches prometheusNS and ruleNS, not arbitraryNS.
-		err := framework.CreatePrometheusOperator(operatorNS, *opImage, []string{prometheusNS, ruleNS})
+		_, err := framework.CreatePrometheusOperator(operatorNS, *opImage, []string{prometheusNS, ruleNS}, nil, false)
 		if err != nil {
 			t.Fatal(err)
 		}
@@ -1714,3 +1766,12 @@ type prometheusAlertmanagerAPIResponse struct {
 	Status string                 `json:"status"`
 	Data   *alertmanagerDiscovery `json:"data"`
 }
+
+func gzipConfig(buf *bytes.Buffer, conf []byte) error {
+	w := gzip.NewWriter(buf)
+	defer w.Close()
+	if _, err := w.Write(conf); err != nil {
+		return err
+	}
+	return nil
+}
